//
// Simple rain screensaver for Win32
//
// Author: Alex V. Boreskoff
//
// Application uses xOffs and yOffs tables for refraction.
//

#include	<windows.h>
#include	<scrnsave.h>
#include	"OffsTable.h"

#define	MAX_DROPS	100

class	Surface;

int	size = 20;						// size of drop
int	width, height;					// screen size

class	Drop
{
	int		x, y;					// coordinates

public:
	Drop ( int theX, int theY )
	{
		x = theX;
		y = theY;
	}

	int	getX () const
	{
		return x;
	}

	int	getY () const
	{
		return y;
	}

	void	draw    ( HDC hDc, Surface * src, Surface * temp, Surface * out ) const;
	void	animate ();
};

class	Surface
{
protected:
	int				width;					// size of surface
	int				height;
	int				bytesPerLine;			// size of single scan line
	byte          * bits;					// image
	HBITMAP			handle;
	LPBITMAPINFO	bmi;

public:
	Surface  ( int theWidth, int theHeight );
	~Surface ();

	virtual	bool	isOk () const
	{
		return width > 0 && height > 0 && bits != NULL;
	}

	HBITMAP	getHandle () const
	{
		return handle;
	}

	int	getWidth () const
	{
		return width;
	}

	int	getHeight () const
	{
		return height;
	}

	byte * getPixelAddr ( int x, int y ) const
	{
		return bits + y * bytesPerLine + x * 3;
	}

	int	getBytesPerLine () const
	{
		return bytesPerLine;
	}

protected:
	bool	build ();
};

///////////////////////////////////////////////////////////////////////////////

/////////////////////////////////// Here go the methods ///////////////////////

void	Drop :: draw ( HDC hDc, Surface * src, Surface * temp, Surface * out ) const
{
								// get corresponding area into temp surface
	HDC		hTempDc  = CreateCompatibleDC ( hDc );
	HDC		hTemp2Dc = CreateCompatibleDC ( hDc );
	HGDIOBJ	saveBmp = SelectObject       ( hTempDc,  temp -> getHandle () );
	HGDIOBJ save2   = SelectObject       ( hTemp2Dc, src  -> getHandle () );

	BitBlt ( hTempDc, 0, 0, size, size, hTemp2Dc, x, y, SRCCOPY );

	SelectObject ( hTempDc,  saveBmp );
	SelectObject ( hTemp2Dc, save2   );
	DeleteDC     ( hTempDc           );
	DeleteDC     ( hTemp2Dc          );

								// now draw the drop
	for ( int iy = 0; iy < size; iy++ )
	{
		if ( y + iy >= height )
			break;

		byte * ptr1 = out -> getPixelAddr ( x, y + iy );

		for ( int ix = 0; ix < size; ix++ )
		{
			if ( ix + size >= width )
				break;

			byte * ptr2 = temp -> getPixelAddr ( xOffs [iy][ix], yOffs [ix][iy] );

			if ( x + ix < width && x + xOffs [iy][ix] < width && y + yOffs [ix][iy] < height )
			{
				ptr1 [3*ix  ] = ptr2 [0];
				ptr1 [3*ix+1] = ptr2 [1];
				ptr1 [3*ix+2] = ptr2 [2];
			}
		}
	}
}

void	Drop :: animate ()
{
	int	rnd1 = (rand () % 5);
	int	rnd2 = (rand () % 3) - 1;

	x += rnd2;
	y += 1 + rnd1;
}

/////////////////////////////////// Surface methods ///////////////////////////////

Surface :: Surface ( int theWidth, int theHeight )
{
	width           = theWidth;
	height          = theHeight;
	handle          = NULL;

	build ();
}

Surface :: ~Surface ()
{
	if ( handle != NULL )
		DeleteObject ( handle );

	if ( bmi != NULL )
		free ( bmi );
}

bool	Surface :: build ()
{
	int	bmiSize = sizeof ( BITMAPINFOHEADER) + 256 * sizeof ( RGBQUAD );

	bits = NULL;
	bmi  = (LPBITMAPINFO) malloc ( bmiSize );

	if ( bmi == NULL )
		return false;

	memset ( bmi, '\0', bmiSize );

									// useless without compression
	bmi -> bmiHeader.biSizeImage     = 0;
	bmi -> bmiHeader.biXPelsPerMeter = 0;
	bmi -> bmiHeader.biYPelsPerMeter = 0;
	bmi -> bmiHeader.biClrUsed       = 0;
	bmi -> bmiHeader.biClrImportant  = 0;

									// self-explanatory params
	bmi -> bmiHeader.biSize        = sizeof ( BITMAPINFOHEADER );
	bmi -> bmiHeader.biWidth       = width;
	bmi -> bmiHeader.biHeight      = -height;		// upside down
	bmi -> bmiHeader.biPlanes      = 1;
	bmi -> bmiHeader.biBitCount    = 24;
	bmi -> bmiHeader.biCompression = BI_RGB;

	HDC	hdc = :: GetDC ( NULL );		// get screen dc

	handle = :: CreateDIBSection ( hdc, bmi, DIB_RGB_COLORS, (void **) &bits, NULL, 0 );

	:: ReleaseDC ( NULL, hdc );

	bytesPerLine = 3 * width;

										// do DWORD alignment
	int	delta = bytesPerLine & 3;

	if ( delta != 0 )
		bytesPerLine += 4 - delta;

	return handle != NULL;
}

///////////////////////////////////////////////////////////////////////////////////

Drop    * drops [MAX_DROPS];
int		  numDrops = 30;
Surface * src;								// source (with initial screen image)
Surface * temp;								// temporary surface (size*size)
Surface * out;								// surface with resulting image
UINT	  timerId;

void	newDrop ( int index )
{
	delete drops [index];

	int	x = 10 + (rand () % (width - 20));
	int	y = rand () % 300;

	drops [index] = new Drop ( x, y );
}

void	draw ( HWND hWnd )
{
	HDC	hDc = GetWindowDC ( hWnd );

	{
		HDC		hTempDc = CreateCompatibleDC ( hDc );
		HGDIOBJ	saveBmp = SelectObject       ( hTempDc, out -> getHandle () );
		HGDIOBJ save2   = SelectObject       ( hDc,     src -> getHandle () );

		BitBlt ( hTempDc, 0, 0, width, height, hDc, 0, 0, SRCCOPY );

		SelectObject ( hTempDc, saveBmp );
		SelectObject ( hDc,     save2   );
		DeleteDC     ( hTempDc );
	}

	for ( int i = 0; i < numDrops; i++ )
	{
		drops [i] -> draw    ( hDc, src, temp, out );
		drops [i] -> animate ();

		if ( drops [i] -> getY () >= height )
			newDrop ( i );
	}

	HDC		hTempDc = CreateCompatibleDC ( hDc );
	HGDIOBJ	saveBmp = SelectObject       ( hTempDc, out -> getHandle () );

	BitBlt ( hDc, 0, 0, width, height, hTempDc, 0, 0, SRCCOPY );

	SelectObject ( hTempDc, saveBmp );
	DeleteDC     ( hTempDc );
	ReleaseDC    ( hWnd, hDc );
}

LRESULT WINAPI ScreenSaverProc ( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HDC		hDc;
	HDC		hTempDc;
	HGDIOBJ	saveBmp;
	RECT	r;
	int		i;

	switch ( message )
    {
		case WM_CREATE:
			GetClientRect  ( hWnd, &r );

			width   = r.right - r.left + 1;
			height  = r.bottom - r.top + 1;
			src     = new Surface ( width, height );
			out     = new Surface ( width, height );
			temp    = new Surface ( size,  size   );
			timerId = SetTimer    ( hWnd, 1, 10, NULL );

			for ( i = 0; i < MAX_DROPS; i++ )
				drops [i] = NULL;

			for ( i = 0; i < numDrops; i++ )
				newDrop ( i );

			hDc     = GetWindowDC ( NULL );
			hTempDc = CreateCompatibleDC ( hDc );
			saveBmp = SelectObject       ( hTempDc, src -> getHandle () );

			BitBlt       ( hTempDc, 0, 0, width, width, hDc, 0, 0, SRCCOPY );
			SelectObject ( hTempDc, saveBmp );
			DeleteDC     ( hTempDc );
			ReleaseDC    ( NULL, hDc );
            break;

		case WM_TIMER:
			draw ( hWnd );
            break;


		case WM_DESTROY:
				delete src;
				delete out;
				delete temp;

				for ( i = 0; i < numDrops; i++ )
					delete drops [i];

                break;

		case WM_ERASEBKGND:
				return 1;

        default:
			return DefScreenSaverProc ( hWnd, message, wParam, lParam );

	}

    return 0;
}

BOOL WINAPI ScreenSaverConfigureDialog ( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam )
{
    switch ( message )
    {
        case WM_COMMAND:
			switch ( LOWORD ( wParam ) )
            {
                case IDOK:  // Exit the dialog
					EndDialog ( hDlg, 42 );
					break;
            }
			break;

        default:
		    return FALSE;
			break;
    }

	return TRUE;
}

BOOL WINAPI RegisterDialogClasses ( HANDLE hInst )
{
	return TRUE;
}

